"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const crypto_1 = require("crypto");
const path_1 = require("path");
const fs_1 = require("fs");
const html_webpack_plugin_1 = require("html-webpack-plugin");
const terser_1 = require("terser");
// @ts-ignore
const core_js_builder_1 = __importDefault(require("core-js-builder"));
class AutoPolyfillsWebpackPlugin {
    constructor(opts = {}) {
        var _a, _b, _c, _d, _e;
        this.opts = opts;
        (_a = this.opts).cwd || (_a.cwd = process.cwd());
        (_b = this.opts).browserslistPath || (_b.browserslistPath = (0, path_1.join)(this.opts.cwd, AutoPolyfillsWebpackPlugin.defaultBrowserslistPath));
        if (!(0, fs_1.existsSync)(this.opts.browserslistPath)) {
            throw new Error(`browserslist file not find,please check if it has been placed in correct path.`);
        }
        (_c = this.opts).customPrefix || (_c.customPrefix = 'dt');
        this.strBeforeHash = `${this.opts.customPrefix}-polyfills-`;
        (_d = this.opts).excludeCoreJsModules || (_d.excludeCoreJsModules = []);
        (_e = this.opts).hashType || (_e.hashType = 'md5');
        this.opts.longTermCache =
            typeof this.opts.longTermCache === 'boolean' && !this.opts.longTermCache
                ? false
                : true;
    }
    apply(compiler) {
        if (compiler.options.mode === 'development' ||
            process.env.NODE_ENV === 'development') {
            return;
        }
        compiler.hooks.beforeRun.tapPromise(AutoPolyfillsWebpackPlugin.name, async () => {
            const polyfillsFiles = this.getDirPolyfillsFiles();
            const targetHashContent = JSON.stringify({
                coreJsVersion: this.coreJsVersion,
                browserslistQuery: this.browserslistQuery,
            });
            this.targetHash = this.getHashResult(targetHashContent);
            // polyfills file name should like `dt-polyfills-${hash}.js`
            this.polyfillFilename = `${this.strBeforeHash}${this.targetHash}.js`;
            let polyfillExist = false;
            polyfillsFiles.forEach(file => {
                const baseName = (0, path_1.basename)(file, '.js');
                const [_, originHash] = baseName.split(this.strBeforeHash);
                if (originHash === this.targetHash) {
                    polyfillExist = true;
                }
                else {
                    (0, fs_1.rmSync)((0, path_1.join)(this.opts.cwd, file));
                }
            });
            if (!polyfillExist) {
                await this.generatePolyfillEntry();
            }
        });
        compiler.hooks.compilation.tap(AutoPolyfillsWebpackPlugin.name, compilation => {
            // html-webpack-plugin hook
            const htmlWebpackHooks = (0, html_webpack_plugin_1.getHooks)(compilation);
            const htmlWebpackHookName = `${AutoPolyfillsWebpackPlugin.name}-html-plugin`;
            htmlWebpackHooks.beforeAssetTagGeneration.tapAsync(htmlWebpackHookName, (data, callback) => {
                const { js: originJsAssets, publicPath } = data.assets;
                originJsAssets.unshift([publicPath, this.polyfillFilename].join(''));
                callback(null, data);
            });
        });
        compiler.hooks.afterEmit.tapPromise(AutoPolyfillsWebpackPlugin.name, async () => {
            const polyfillPath = (0, path_1.join)(this.opts.cwd, this.polyfillFilename);
            (0, fs_1.copyFileSync)(polyfillPath, (0, path_1.join)(compiler.options.output.path, this.polyfillFilename));
            if (!this.opts.longTermCache) {
                (0, fs_1.rmSync)(polyfillPath);
            }
        });
    }
    getDirPolyfillsFiles() {
        return (0, fs_1.readdirSync)(this.opts.cwd).filter(path => path.includes(this.strBeforeHash) &&
            !(0, fs_1.statSync)((0, path_1.join)(this.opts.cwd, path)).isDirectory());
    }
    async generatePolyfillEntry() {
        const filename = (0, path_1.join)(this.opts.cwd, this.polyfillFilename);
        try {
            const bundleContent = await (0, core_js_builder_1.default)({
                targets: this.browserslistQuery,
                exclude: this.opts.excludeCoreJsModules,
            });
            const { code } = await (0, terser_1.minify)(bundleContent);
            (0, fs_1.writeFileSync)(filename, code);
        }
        catch (ex) {
            throw ex;
        }
    }
    getHashResult(content) {
        const hash = (0, crypto_1.createHash)(this.opts.hashType);
        hash.update(content);
        return hash.digest('hex');
    }
    get coreJsVersion() {
        try {
            const coreJsPathList = require.resolve('core-js').split('/');
            coreJsPathList.pop();
            const coreJsPkg = (0, path_1.join)(coreJsPathList.join('/'), './package.json');
            return require(coreJsPkg).version;
        }
        catch (ex) {
            throw ex;
        }
    }
    get browserslistQuery() {
        try {
            const content = (0, fs_1.readFileSync)(this.opts.browserslistPath, 'utf-8');
            const queryList = content
                .split('\n')
                .map(txt => txt.replace(/#((\s)*(\w)+)*/, '').trim())
                .filter(Boolean);
            return queryList.join(',');
        }
        catch (ex) {
            throw ex;
        }
    }
}
AutoPolyfillsWebpackPlugin.defaultBrowserslistPath = '.browserslistrc';
exports.default = AutoPolyfillsWebpackPlugin;
